LIBRARY test_icd_version_7_with_additional_exports
EXPORTS
    vk_icdGetInstanceProcAddr
    vk_icdNegotiateLoaderICDInterfaceVersion
    vk_icdEnumerateAdapterPhysicalDevices
    vk_icdGetPhysicalDeviceProcAddr
